/**
 * 
 */
package org.processmininguc.configurableprocesses.algorithms;

import java.util.Random;

import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.impl.XEventImpl;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.etm.CentralRegistry;
import org.processmining.plugins.etm.fitness.TreeFitness;
import org.processmining.plugins.etm.fitness.TreeFitnessInfo;
import org.processmining.plugins.etm.fitness.metrics.OverallFitness;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.conversion.ProcessTreeToNAryTree;
import org.processmining.plugins.etm.parameters.ETMParamFactory;
import org.processmining.plugins.etm.ui.plugins.PTQualityCalculator;
import org.processmining.processtree.Block;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.impl.AbstractTask.Manual;
import org.processmininguc.configurableprocesses.parameters.ExhaustiveParameters;
import org.processmininguc.configurableprocesses.parameters.GeneticParameters;
import org.processmininguc.configurableprocesses.parameters.GreedyParameters;

import gnu.trove.iterator.TObjectDoubleIterator;

/**
 * @author mauro
 *
 */
public class ConformanceMetric {

	public static TreeFitness getProcessTreeConformance(PluginContext context, ProcessTree tree, XLog log, XEventClassifier classifier){
		
		//FIXME more robust way to determine classifier if none is given...
		if (classifier == null) {
			classifier = XLogInfoImpl.STANDARD_CLASSIFIER;
			//			classifier = XLogInfoImpl.NAME_CLASSIFIER;
		}
		
		CentralRegistry registry = new CentralRegistry(log, new Random());
		registry.updateEventClassifier(classifier);
		
			//add the event classes of the tree manually
		addAllLeaves(registry.getEventClasses(), tree.getRoot());
		
		ProcessTreeToNAryTree ptToNat = new ProcessTreeToNAryTree(registry.getEventClasses());
		NAryTree naTree = ptToNat.convert(tree);
		
		registry.updateLogDerived();

		OverallFitness of = ETMParamFactory.createStandardOverallFitness(registry);
		of.getFitness(naTree, null);

		TreeFitness fitness = registry.getFitness(naTree);

		return fitness;
	}

	//Method borrowed from org.processmining.plugins.inductiveVisualMiner.alignment.AlignmentETM revision 16927
	public static void addAllLeaves(XEventClasses classes, Node node) {
		if (node instanceof Manual) {
			XEvent event = new XEventImpl();
			XConceptExtension.instance().assignName(event, node.getName());
			classes.register(event);
		} else if (node instanceof Block) {
			for (Node child : ((Block) node).getChildren()) {
				addAllLeaves(classes, child);
			}
		}
		classes.harmonizeIndices();
	}	
	
	
	//obtaining the overall fitness metric from Buij's plugin
	//To be used by Greedy strategy
	public static Double getNAryTreeConformance(PluginContext context, ProcessTree tree, XLog log, GreedyParameters parameters, XEventClassifier classifier, final ConfigurableProcessesResult confProcResults ){
		PTQualityCalculator PTQuality = new PTQualityCalculator();
		TreeFitness fitnessMetric = PTQuality.calculatePTQuality(context, log, tree, classifier);
		confProcResults.setOveralFitnessMetric(fitnessMetric);
		confProcResults.setParamFitness(parameters.getFitnessWeigth());
		confProcResults.setParamPrecision(parameters.getPrecisionWeigth());
		confProcResults.setParamGeneralization(parameters.getGeneralizationWeigth());
		confProcResults.setParamSimplicity(parameters.getSimplicityWeigth());
		double fitness = 0.0, denominator, numerator;
		double overallFitness = 0.0;
        double precision=0.0, simplicity=0.0, generalization=0.0, conformance=0.0, bestConformance=0.0;

		TObjectDoubleIterator<TreeFitnessInfo> it = fitnessMetric.fitnessValues.iterator();
		
		while (it.hasNext()) {
			it.advance();
			if (it.key().toString().equals("Fr"))
				fitness = it.value();
			if (it.key().toString().equals("Of"))
				overallFitness = it.value();
			if (it.key().toString().equals("Su"))
				simplicity = it.value();
			if (it.key().toString().equals("Pe"))
				precision = it.value();
			if (it.key().toString().equals("Gv"))
				generalization = it.value();
		}        	
		confProcResults.setFitness(fitness);
		confProcResults.setPrecision(precision);
		confProcResults.setGeneralization(generalization);
		confProcResults.setSimplcity(simplicity);
		
		numerator = fitness*parameters.getFitnessWeigth()/((double) 100) + precision*parameters.getPrecisionWeigth()/((double) 100) + 
				  generalization*parameters.getGeneralizationWeigth()/((double) 100) + simplicity*parameters.getSimplicityWeigth()/((double) 100);
		denominator = (parameters.getFitnessWeigth()/((double) 100) + parameters.getPrecisionWeigth()/((double) 100) + parameters.getGeneralizationWeigth()/((double) 100) + parameters.getSimplicityWeigth()/((double) 100));

		conformance = numerator/denominator;

		return conformance;		
	}

	//To be used by Exhaustive strategy
	public static Double getNAryTreeConformance(PluginContext context, ProcessTree tree, XLog log, ExhaustiveParameters parameters, XEventClassifier classifier, final ConfigurableProcessesResult confProcResults ){
		PTQualityCalculator PTQuality = new PTQualityCalculator();
		TreeFitness fitnessMetric = PTQuality.calculatePTQuality(context, log, tree, classifier);
		confProcResults.setOveralFitnessMetric(fitnessMetric);
		confProcResults.setParamFitness(parameters.getFitnessWeigth());
		confProcResults.setParamPrecision(parameters.getPrecisionWeigth());
		confProcResults.setParamGeneralization(parameters.getGeneralizationWeigth());
		confProcResults.setParamSimplicity(parameters.getSimplicityWeigth());
		double fitness = 0.0, denominator, numerator;
		double overallFitness = 0.0;
        double precision=0.0, simplicity=0.0, generalization=0.0, conformance=0.0, bestConformance=0.0;

		TObjectDoubleIterator<TreeFitnessInfo> it = fitnessMetric.fitnessValues.iterator();
		
		while (it.hasNext()) {
			it.advance();
			if (it.key().toString().equals("Fr"))
				fitness = it.value();
			if (it.key().toString().equals("Of"))
				overallFitness = it.value();
			if (it.key().toString().equals("Su"))
				simplicity = it.value();
			if (it.key().toString().equals("Pe"))
				precision = it.value();
			if (it.key().toString().equals("Gv"))
				generalization = it.value();
		}        	
		confProcResults.setFitness(fitness);
		confProcResults.setPrecision(precision);
		confProcResults.setGeneralization(generalization);
		confProcResults.setSimplcity(simplicity);
		
		numerator = fitness*parameters.getFitnessWeigth()/((double) 100) + precision*parameters.getPrecisionWeigth()/((double) 100) + 
				  generalization*parameters.getGeneralizationWeigth()/((double) 100) + simplicity*parameters.getSimplicityWeigth()/((double) 100);
		denominator = (parameters.getFitnessWeigth()/((double) 100) + parameters.getPrecisionWeigth()/((double) 100) + parameters.getGeneralizationWeigth()/((double) 100) + parameters.getSimplicityWeigth()/((double) 100));

		conformance = numerator/denominator;

		return conformance;		
	}	
	
	//To be used by Genetic strategy
	public static Double getNAryTreeConformance(PluginContext context, ProcessTree tree, XLog log, GeneticParameters parameters, XEventClassifier classifier, final ConfigurableProcessesResult confProcResults ){
		PTQualityCalculator PTQuality = new PTQualityCalculator();
		TreeFitness fitnessMetric = PTQuality.calculatePTQuality(context, log, tree, classifier);
		confProcResults.setOveralFitnessMetric(fitnessMetric);
		confProcResults.setParamFitness(parameters.getFitnessWeigth());
		confProcResults.setParamPrecision(parameters.getPrecisionWeigth());
		confProcResults.setParamGeneralization(parameters.getGeneralizationWeigth());
		confProcResults.setParamSimplicity(parameters.getSimplicityWeigth());
		double fitness = 0.0, denominator, numerator;
		double overallFitness = 0.0;
        double precision=0.0, simplicity=0.0, generalization=0.0, conformance=0.0, bestConformance=0.0;

		TObjectDoubleIterator<TreeFitnessInfo> it = fitnessMetric.fitnessValues.iterator();
		
		while (it.hasNext()) {
			it.advance();
			if (it.key().toString().equals("Fr"))
				fitness = it.value();
			if (it.key().toString().equals("Of"))
				overallFitness = it.value();
			if (it.key().toString().equals("Su"))
				simplicity = it.value();
			if (it.key().toString().equals("Pe"))
				precision = it.value();
			if (it.key().toString().equals("Gv"))
				generalization = it.value();
		}        	
		confProcResults.setFitness(fitness);
		confProcResults.setPrecision(precision);
		confProcResults.setGeneralization(generalization);
		confProcResults.setSimplcity(simplicity);
		
		numerator = fitness*parameters.getFitnessWeigth()/((double) 100) + precision*parameters.getPrecisionWeigth()/((double) 100) + 
				  generalization*parameters.getGeneralizationWeigth()/((double) 100) + simplicity*parameters.getSimplicityWeigth()/((double) 100);
		denominator = (parameters.getFitnessWeigth()/((double) 100) + parameters.getPrecisionWeigth()/((double) 100) + parameters.getGeneralizationWeigth()/((double) 100) + parameters.getSimplicityWeigth()/((double) 100));

		conformance = numerator/denominator;

		return conformance;		
	}		
	
}
