package org.processmininguc.configurableprocesses.algorithms;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.UUID;

import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.etm.CentralRegistry;
import org.processmining.plugins.etm.fitness.TreeFitness;
import org.processmining.plugins.etm.fitness.TreeFitnessAbstract;
import org.processmining.plugins.etm.fitness.TreeFitnessInfo;
import org.processmining.plugins.etm.model.narytree.Configuration;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.conversion.NAryTreeToProcessTree;
import org.processmining.plugins.etm.model.narytree.conversion.ProcessTreeToNAryTree;
import org.processmining.plugins.etm.ui.plugins.PTQualityCalculator;
import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmininguc.configurableprocesses.parameters.ExhaustiveParameters;
import org.processmininguc.configurableprocesses.tests.Pair;

import gnu.trove.iterator.TObjectDoubleIterator;

public class ExhaustiveAlgorithm {

	/**
	 * The method that implements your algorithm.
	 * 
	 * Note that this method only uses the boolean which is stored in the parameters.
	 * Nevertheless, it could have used the integer and/or the String as well.
	 * 
	 * @param context The context where to run in.
	 * @param input1 The first input.
	 * @param input2 The second input.
	 * @param parameters The parameters to use.
	 * @return The output.
	 */
	public ProcessTree apply(PluginContext context, ProcessTree tree, XLog log, ExhaustiveParameters parameters,
			final ConfigurableProcessesResult confProcResults) {
		/**
		 * Put your algorithm here, which computes an output form the inputs provided the parameters.
		 */
		long startTime = System.currentTimeMillis();

		ProcessTree bestProcessTree = exhasutiveAlgorithm(context,tree, log, parameters,confProcResults);
		
		long endTime   = System.currentTimeMillis();
		String finalTime = showElapsedTime(startTime,endTime);
		
		confProcResults.setTimeElapsed(finalTime);
	    return bestProcessTree;
	}

	/**
	 * The method that implements your algorithm.
	 * 
	 * Note that this method only uses the boolean which is stored in the parameters.
	 * Nevertheless, it could have used the integer and/or the String as well.
	 * @param parameters 
	 * 
	 * @param Process Tree.
	 * @param Log.
	 * @return A ProcessTree that represent the best configuration.
	 */
	public static ProcessTree exhasutiveAlgorithm(PluginContext context, ProcessTree tree, XLog log, ExhaustiveParameters parameters,
			final ConfigurableProcessesResult confProcResults){
		CentralRegistry registry = new CentralRegistry(log, XLogInfoImpl.NAME_CLASSIFIER, new Random());

		PTQualityCalculator PTQuality = new PTQualityCalculator();
		TreeFitness fitnessMetric = null;
		
		Map<String,String> finalConfiguration = new HashMap<String,String>(); //this holds the final configuration
		
		ProcessTreeToNAryTree converter = new ProcessTreeToNAryTree(registry.getEventClasses());
		NAryTree simpleNAryTree = converter.convert(tree);
					
		List<UUID> uuidsNodeList = new ArrayList<UUID>();	//simple list of UUID from nodes
		uuidsNodeList = treeUUIDList(tree.getRoot(), uuidsNodeList);	//pre-order list of tree nodes 
		
		Map<UUID,List<Byte>> uuidsConfsMap = new HashMap<UUID,List<Byte>>();
		List<List<Pair<UUID,Byte>>> pairList = new ArrayList<List<Pair<UUID,Byte>>>();
		
		for (Edge edge :tree.getEdges()){
			List<Byte> bytelist = new ArrayList<Byte>();
			if (edge.isBlockable()){
				bytelist.add(Configuration.BLOCKED);
			}
			if(edge.isHideable()){				
				bytelist.add(Configuration.HIDDEN);
			}
			if(edge.isBlockable() || edge.isHideable()){				
				bytelist.add(Configuration.NOTCONFIGURED);
				uuidsConfsMap.put(tree.getEdge(edge.getID()).getTarget().getID(),bytelist);
			}
		}
		
		//Create list of pairs<UUID,Byte> ... id and configuration
    	Pair<UUID, Byte> p = new Pair(null,null);
        for (Map.Entry<UUID, List<Byte>> entry : uuidsConfsMap.entrySet()) {
        	List<Pair<UUID,Byte>> temp = new ArrayList<Pair<UUID,Byte>>();
            for (Byte ent : entry.getValue()){
            	p = new Pair(entry.getKey(),ent);
            	temp.add(p);
            }
        	pairList.add(temp);
        }			

        //combineAllCases(uuidsConfsMap);
        List<List<Pair<UUID,Byte>>> cartProdList = new ArrayList<List<Pair<UUID,Byte>>>(cartesianProduct(pairList));
        TreeFitnessAbstract of = null;
        ProcessTree convertedTree=null, bestTree=null;
        Configuration c=null, bestConfiguration=null;
        int i=0;
        double bestFitness=0.0, bestOverallFitness=0.0, fitness=0.0, overallFitness=0.0;
        double precision=0.0, simplicity=0.0, generalization=0.0, conformance=0.0, bestConformance=0.0;
        double numerator, denominator;
        
        List<Pair<UUID,Byte>> bestconfList = new ArrayList<Pair<UUID,Byte>>();
        
        for(List<Pair<UUID,Byte>> confList : cartProdList){
        	if (i%100==0)
        		System.out.println("******* Proccessing configuration "+(i+1)+" *******");
        	c = new Configuration(simpleNAryTree.size());

        	for(Pair<UUID,Byte> comb : confList){ //cartProdList.get(0) o confList
        		if (comb.getR().equals(Configuration.BLOCKED)){
        			c.conf[uuidsNodeList.indexOf(comb.getL())]=Configuration.BLOCKED;
        		}else if(comb.getR().equals(Configuration.HIDDEN)){
        			c.conf[uuidsNodeList.indexOf(comb.getL())]=Configuration.HIDDEN;
        		}else if(comb.getR().equals(Configuration.NOTCONFIGURED)){
        			c.conf[uuidsNodeList.indexOf(comb.getL())]=Configuration.NOTCONFIGURED;
        		}
        	}
        	
        	simpleNAryTree.addConfiguration(c);
        	NAryTree res = simpleNAryTree.applyConfiguration(0);
    		//of = ETMParamFactory.createStandardOverallFitness(registry);

        	convertedTree =  NAryTreeToProcessTree.convert(res, registry.getEventClasses());
   	
        	//Calling external overallFitness
        	fitnessMetric = PTQuality.calculatePTQuality(context, log, convertedTree, XLogInfoImpl.NAME_CLASSIFIER);
        	//fitnessMetric = ConformanceMetric.getProcessTreeConformance(context, convertedTree, log, XLogInfoImpl.NAME_CLASSIFIER);
        	//{Fr=1.0,Pe=1.0,Gv=0.8834314575050762,Of=0.9931430269120632,Su=1.0}
        	
        	TreeFitness fit = registry.getFitness(res);
        	//System.out.println("res fitness : "+fit);
        	
        	TObjectDoubleIterator<TreeFitnessInfo> it = fitnessMetric.fitnessValues.iterator();
			while (it.hasNext()) {
				it.advance();
				if (it.key().toString().equals("Fr"))
					fitness = it.value();
				if (it.key().toString().equals("Of"))
					overallFitness = it.value();
				if (it.key().toString().equals("Su"))
					simplicity = it.value();
				if (it.key().toString().equals("Pe"))
					precision = it.value();
				if (it.key().toString().equals("Gv"))
					generalization = it.value();
			}        	
			i++;
        	simpleNAryTree.removeConfiguration(0);
        	
    		numerator = fitness*parameters.getFitnessWeigth()/((double) 100) + precision*parameters.getPrecisionWeigth()/((double) 100) + 
  				  generalization*parameters.getGeneralizationWeigth()/((double) 100) + simplicity*parameters.getSimplicityWeigth()/((double) 100);
    		denominator = (parameters.getFitnessWeigth()/((double) 100) + parameters.getPrecisionWeigth()/((double) 100) + parameters.getGeneralizationWeigth()/((double) 100) + parameters.getSimplicityWeigth()/((double) 100));

    		conformance = numerator/denominator;
        	
        	//first we compare the fitness metric. In case the fitness are equal, we compare the overall fitness
        	if (conformance > bestConformance){	//fitness > bestFitness
        		bestConformance = conformance; 	//bestFitness = fitness;
        		bestOverallFitness = overallFitness;
        		bestTree = NAryTreeToProcessTree.convert(res, registry.getEventClasses());
        		bestConfiguration = c;
        		bestconfList = confList;
        	}
        	
        }
        //put the node and its configurator
        int nConfigurableNodes = 0;
        for(Pair<UUID,Byte> comb : bestconfList){ //cartProdList.get(0) o confList
            Node node = tree.getNode(comb.getL());
    		Block block = node.getIncomingEdges().get(0).getSource();
        	finalConfiguration.put(node.toString(),null);
    		if (comb.getR().equals(Configuration.BLOCKED)){
    			finalConfiguration.put(tree.getNode(comb.getL()).getName(),"B");
    			nConfigurableNodes++;
    			//c.conf[uuidsNodeList.indexOf(comb.getL())]=Configuration.BLOCKED;
    		}else if(comb.getR().equals(Configuration.HIDDEN)){
    			finalConfiguration.put(node.toString(),"H");
    			nConfigurableNodes++;
    			//c.conf[uuidsNodeList.indexOf(comb.getL())]=Configuration.HIDDEN;
    		}else if(comb.getR().equals(Configuration.NOTCONFIGURED)){
    			finalConfiguration.put(node.toString(),"E");
    			nConfigurableNodes++;
    			//c.conf[uuidsNodeList.indexOf(comb.getL())]=Configuration.NOTCONFIGURED;
    		}
    	}
        confProcResults.setFinalConf(finalConfiguration);
        confProcResults.setConfigurationSize(nConfigurableNodes);
        //confProcResults.setFinalConf(bestConfiguration);
        Double finalconf = ConformanceMetric.getNAryTreeConformance(context,tree,log,parameters,XLogInfoImpl.NAME_CLASSIFIER,confProcResults);
        confProcResults.setConformance(finalconf);
        confProcResults.setBestConfiguration(bestConfiguration);
        System.out.println("==== Show Best Configuration ====");
        System.out.println(" best Tree : "+bestTree+
        					"\n best Configuration : "+bestConfiguration+
        					"\n best conformance : "+bestConformance+
        					"\n best overall fitness : "+bestOverallFitness);
         
		return bestTree;
	}	
	
	public static List<UUID> treeUUIDList(Node node, List<UUID> uuidList){
		if (node.isLeaf()){
			uuidList.add(node.getID());
			return uuidList;
		}else{
			uuidList.add(node.getID());
			Block b = (Block) node;
			for (Node nod: b.getChildren()){
				treeUUIDList(nod, uuidList);
			}
		}
		return uuidList;
	}
	
	public static List<List<Pair<UUID,Byte>>> cartesianProduct(List<List<Pair<UUID,Byte>>> pairLists) {
		List<List<Pair<UUID,Byte>>> resultLists = new ArrayList<List<Pair<UUID,Byte>>>();
	    if (pairLists.size() == 0) {
	        resultLists.add(new ArrayList<Pair<UUID,Byte>>());
	        return resultLists;
	    } else {
	    	List<Pair<UUID,Byte>> firstList = pairLists.get(0);
	        List<List<Pair<UUID,Byte>>> remainingLists = cartesianProduct(pairLists.subList(1, pairLists.size()));
	        for (Pair<UUID,Byte> condition : firstList) {
	            for (List<Pair<UUID,Byte>> remainingList : remainingLists) {
	                ArrayList<Pair<UUID,Byte>> resultList = new ArrayList<Pair<UUID,Byte>>();
	                resultList.add(condition);
	                resultList.addAll(remainingList);
	                resultLists.add(resultList);
	            }
	        }
	    }
	    return resultLists;
	}	
	
	
	private String showElapsedTime(long startTime, long endTime) {
		// TODO Auto-generated method stub
		long time = endTime - startTime;
		long milli  = time % 1000;
		long second = (time / 1000) % 60;
		long minute = (time / (1000 * 60)) % 60;
		long hour = (time / (1000 * 60 * 60)) % 24;
		String timeFormatted = String.format("%02d:%02d:%02d:%d", hour, minute, second, milli);
		System.out.println("Time elapsed : "+timeFormatted);
		return timeFormatted;
	}

}
