/**
 * 
 */
package org.processmininguc.configurableprocesses.algorithms;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.UUID;

import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XLog;
import org.jgap.Chromosome;
import org.jgap.Configuration;
import org.jgap.FitnessFunction;
import org.jgap.Gene;
import org.jgap.Genotype;
import org.jgap.IChromosome;
import org.jgap.InvalidConfigurationException;
import org.jgap.impl.DefaultConfiguration;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.etm.CentralRegistry;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.conversion.NAryTreeToProcessTree;
import org.processmining.plugins.etm.model.narytree.conversion.ProcessTreeToNAryTree;
import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmininguc.configurableprocesses.parameters.GeneticParameters;

/**
 * @author mauro
 *
 */
public class GeneticAlgorithm {

	/**
	 * The method that implements your algorithm.
	 * 
	 * Note that this metxt where to run in.
	 * @param input1 The first input.
	 * @param input2 The second input.
	 * @param parameters The parameters to use.
	 * @return The output.
	 */	
	
	//General variables
	private PluginContext context;
	private XLog log;
	private ProcessTree tree; 
	private org.jgap.Configuration JGAPConfiguration;
	private GeneticParameters parameters;
	private CentralRegistry registry;

	private Map<Integer,UUID> fromGenePositionToNodeId;
	private Map<UUID,Integer> fromNodeIdToGenePosition;	
	private Map<UUID,List<Byte>> allUUIDsNodeConfsMap = new HashMap<UUID,List<Byte>>();
	private List<UUID> uuidsNodeList = new ArrayList<UUID>();	//simple list of UUID from nodes

	
	public ProcessTree apply(PluginContext context, ProcessTree tree, XLog log, GeneticParameters parameters, 
			final ConfigurableProcessesResult confProcResults) throws Exception {
		/**
		 * Put your algorithm here, which computes an output form the inputs provided the parameters.
		 */
		long startTime = System.currentTimeMillis();
		
		this.context = context;
		this.log = log;
		this.parameters = parameters;
		this.tree = tree;
		this.registry = new CentralRegistry(log, XLogInfoImpl.NAME_CLASSIFIER, new Random());

		//List of all general configurations 
		allUUIDsNodeConfsMap = getAllConfigurations(tree, allUUIDsNodeConfsMap);
		uuidsNodeList = treeUUIDList(tree.getRoot(), uuidsNodeList);	//pre-order list of tree nodes 
		
		Configuration.reset();
		createConfiguration();
		createFitnessFunction();
		createChromosome();
		
		Genotype population = createInitialPopulation();
		
		IChromosome bestSolution = null;
		Map<Double,ProcessTree> finalSolution = new LinkedHashMap<Double,ProcessTree>();
		
		for( int i = 0; i < parameters.getMaxAllowedEvolutions(); i++){
			System.out.println("******* evol : "+i+" *******");
			population.evolve();
			bestSolution = population.getFittestChromosome();
			System.out.println("******* best_solution : "+bestSolution.getFitnessValue()+" *******");
		}
		
		bestSolution = population.getFittestChromosome();
		//System.out.println("aqui lo age= "+bestSolution.getConfiguration());
		
		System.out.println(bestSolution.getFitnessValue());
		ProcessTree bestProcessTree = convertFromChromosomeToConfiguredProcessTree(bestSolution,confProcResults);
		System.out.println(bestSolution.getConfiguration());
		
		long endTime   = System.currentTimeMillis();
		String finalTime = showTimeElapsed(startTime,endTime);
		confProcResults.setTimeElapsed(finalTime);
	    return bestProcessTree;
	}

	public static ProcessTree geneticAlgorithm(PluginContext context, ProcessTree tree, XLog log){
        return tree;
	}

	private Genotype createInitialPopulation() throws Exception{
		return Genotype.randomInitialGenotype(JGAPConfiguration);
	}	
	
	private void createConfiguration() throws Exception {
		//TODO Set the all the options of the configuration from the parameters
		JGAPConfiguration = new DefaultConfiguration();
		JGAPConfiguration.setPreservFittestIndividual(true);
		JGAPConfiguration.setKeepPopulationSizeConstant(false);
		JGAPConfiguration.setPopulationSize(parameters.getPopulationSize());
	}	
	
	private void createFitnessFunction() throws Exception{
		FitnessFunction conformanceFitnessFunction = new MaximizingConformanceInConfigurableProcessTreeFitnessFunction(context, log, tree, parameters,  this);
		JGAPConfiguration.setFitnessFunction(conformanceFitnessFunction);
	}	

	private void createChromosome() throws InvalidConfigurationException{
		//List<UUID> uuidsNodeList = null;	//simple list of UUID from nodes
		List<UUID> configurableNodeIds = getAllConfigurationUUID(this.tree);
		
		Gene[] sampleGenes = new Gene[configurableNodeIds.size()];
		
		this.fromGenePositionToNodeId = new LinkedHashMap<Integer,UUID>();
		this.fromNodeIdToGenePosition = new LinkedHashMap<UUID,Integer>();
		
		for(int i=0; i < configurableNodeIds.size(); i++){
			UUID nodeId = configurableNodeIds.get(i);
			this.fromGenePositionToNodeId.put(i, nodeId);
			this.fromNodeIdToGenePosition.put(nodeId, i);
			List<Byte> options = 
					new ArrayList<Byte>(getConfigurationOptions(nodeId));
			sampleGenes[i] = new ProcessTreeConfigurationOptionGene(this.JGAPConfiguration, options);
		}

	    IChromosome sampleChromosome = new Chromosome(JGAPConfiguration, sampleGenes);
	    JGAPConfiguration.setSampleChromosome(sampleChromosome);
	}

	private static List<UUID> getAllConfigurationUUID(ProcessTree tree) {
		// TODO Auto-generated method stub
		List<UUID> confNodeList = new ArrayList<UUID>();
		for (Edge edge :tree.getEdges()){
			if (edge.isBlockable() || edge.isHideable())
				confNodeList.add(tree.getEdge(edge.getID()).getTarget().getID());
		}
		return confNodeList;
	}	
	
	private List<Byte> getConfigurationOptions(UUID nodeId){
		List<Byte> optionsList = allUUIDsNodeConfsMap.get(nodeId);
		return optionsList;
	}

	private static Map<UUID, List<Byte>> getAllConfigurations(ProcessTree tree, Map<UUID, List<Byte>> uuidsNodeConfsMap) {
		// TODO Auto-generated method stub
		for (Edge edge :tree.getEdges()){
			List<Byte> bytelist = new ArrayList<Byte>();
			if (edge.isBlockable())
				bytelist.add(org.processmining.plugins.etm.model.narytree.Configuration.BLOCKED);
			if(edge.isHideable())
				bytelist.add(org.processmining.plugins.etm.model.narytree.Configuration.HIDDEN);
			if(edge.isBlockable() || edge.isHideable()){				
				bytelist.add(org.processmining.plugins.etm.model.narytree.Configuration.NOTCONFIGURED);
				uuidsNodeConfsMap.put(tree.getEdge(edge.getID()).getTarget().getID(),bytelist);
			}
		}
		return uuidsNodeConfsMap;
	}	
	
	protected ProcessTree convertFromChromosomeToConfiguredProcessTree(IChromosome bestSolution) {
		Gene[] genes = bestSolution.getGenes();
		ProcessTree convertedTree = null;
		ProcessTreeToNAryTree converter = new ProcessTreeToNAryTree(registry.getEventClasses());
		NAryTree simpleNAryTree = converter.convert(tree);
		org.processmining.plugins.etm.model.narytree.Configuration c = new org.processmining.plugins.etm.model.narytree.Configuration(simpleNAryTree.size());
		
		for(int i = 0; i < genes.length; i++){
			UUID nodeId = getNodeIdByGenePosition(i);
			ProcessTreeConfigurationOptionGene optionGene = (ProcessTreeConfigurationOptionGene) genes[i];
			Byte value = (Byte) optionGene.getAllele();

    		if (value.equals(org.processmining.plugins.etm.model.narytree.Configuration.BLOCKED)){
    			c.conf[uuidsNodeList.indexOf(nodeId)]=org.processmining.plugins.etm.model.narytree.Configuration.BLOCKED;
    		}else if(value.equals(org.processmining.plugins.etm.model.narytree.Configuration.HIDDEN)){
    			c.conf[uuidsNodeList.indexOf(nodeId)]=org.processmining.plugins.etm.model.narytree.Configuration.HIDDEN;
    		}else if(value.equals(org.processmining.plugins.etm.model.narytree.Configuration.NOTCONFIGURED)){
    			c.conf[uuidsNodeList.indexOf(nodeId)]=org.processmining.plugins.etm.model.narytree.Configuration.NOTCONFIGURED;
    		}          
			
        	simpleNAryTree.addConfiguration(c);
        	NAryTree res = simpleNAryTree.applyConfiguration(0);
        	convertedTree =  NAryTreeToProcessTree.convert(res, registry.getEventClasses());
        	simpleNAryTree.removeConfiguration(0);
		}
		
		return convertedTree;
	}	

	protected ProcessTree convertFromChromosomeToConfiguredProcessTree(IChromosome bestSolution, 
			final ConfigurableProcessesResult confProcResults) {
		Gene[] genes = bestSolution.getGenes();
		ProcessTree convertedTree = null;
		ProcessTreeToNAryTree converter = new ProcessTreeToNAryTree(registry.getEventClasses());
		NAryTree simpleNAryTree = converter.convert(tree);
		org.processmining.plugins.etm.model.narytree.Configuration c = new org.processmining.plugins.etm.model.narytree.Configuration(simpleNAryTree.size());
		Map<String,String> finalConfiguration = new HashMap<String,String>(); //this holds the final configuration
		int nconfigurableNodes = 0;
		for(int i = 0; i < genes.length; i++){
			UUID nodeId = getNodeIdByGenePosition(i);
			ProcessTreeConfigurationOptionGene optionGene = (ProcessTreeConfigurationOptionGene) genes[i];
			Byte value = (Byte) optionGene.getAllele();

    		if (value.equals(org.processmining.plugins.etm.model.narytree.Configuration.BLOCKED)){
    			c.conf[uuidsNodeList.indexOf(nodeId)]=org.processmining.plugins.etm.model.narytree.Configuration.BLOCKED;
    			finalConfiguration.put(tree.getNode(nodeId).getName(),"B");
    			nconfigurableNodes++;
    		}else if(value.equals(org.processmining.plugins.etm.model.narytree.Configuration.HIDDEN)){
    			c.conf[uuidsNodeList.indexOf(nodeId)]=org.processmining.plugins.etm.model.narytree.Configuration.HIDDEN;
    			finalConfiguration.put(tree.getNode(nodeId).toString(),"H");
    			nconfigurableNodes++;
    		}else if(value.equals(org.processmining.plugins.etm.model.narytree.Configuration.NOTCONFIGURED)){
    			c.conf[uuidsNodeList.indexOf(nodeId)]=org.processmining.plugins.etm.model.narytree.Configuration.NOTCONFIGURED;
    			finalConfiguration.put(tree.getNode(nodeId).toString(),"E");
    			nconfigurableNodes++;
    		}          
		
    		simpleNAryTree.addConfiguration(c);
        	NAryTree res = simpleNAryTree.applyConfiguration(0);
        	convertedTree =  NAryTreeToProcessTree.convert(res, registry.getEventClasses());
        	simpleNAryTree.removeConfiguration(0);
		}
		
        Double finalconf = ConformanceMetric.getNAryTreeConformance(context,convertedTree,log,parameters,XLogInfoImpl.NAME_CLASSIFIER,confProcResults);
		confProcResults.setFinalConf(finalConfiguration);
		confProcResults.setConfigurationSize(nconfigurableNodes);
		System.out.println("final configuration double = "+finalconf);
		confProcResults.setConformance(bestSolution.getFitnessValue());
		return convertedTree;
	}
	
	private UUID getNodeIdByGenePosition(int index){
		return this.fromGenePositionToNodeId.get(index);
	}	

	public static List<UUID> treeUUIDList(Node node, List<UUID> uuidList){
		if (node.isLeaf()){
			uuidList.add(node.getID());
			return uuidList;
		}else{
			uuidList.add(node.getID());
			Block b = (Block) node;
			for (Node nod: b.getChildren()){
				treeUUIDList(nod, uuidList);
			}
		}
		return uuidList;
	}

	private String showTimeElapsed(long startTime, long endTime) {
		// TODO Auto-generated method stub
		long time = endTime - startTime;
		long milli  = time % 1000;
		long second = (time / 1000) % 60;
		long minute = (time / (1000 * 60)) % 60;
		long hour = (time / (1000 * 60 * 60)) % 24;
		String timeFormatted = String.format("%02d:%02d:%02d:%d", hour, minute, second, milli);
		System.out.println("Time elapsed : "+timeFormatted);
		return timeFormatted;
	}

	
	
}
