package org.processmininguc.configurableprocesses.algorithms;

import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XLog;
import org.jgap.FitnessFunction;
import org.jgap.IChromosome;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.etm.fitness.TreeFitness;
import org.processmining.plugins.etm.fitness.TreeFitnessInfo;
import org.processmining.plugins.etm.ui.plugins.PTQualityCalculator;
import org.processmining.processtree.ProcessTree;
import org.processmininguc.configurableprocesses.parameters.GeneticParameters;

import gnu.trove.iterator.TObjectDoubleIterator;

public class MaximizingConformanceInConfigurableProcessTreeFitnessFunction extends FitnessFunction {

	private static final long serialVersionUID = -835609389938869422L;
	
	private XLog log;
	private ProcessTree tree;
	private PTQualityCalculator conformancePlugin;
	private PluginContext context;
	private GeneticAlgorithm problem;
	private GeneticParameters parameters;
	
	public MaximizingConformanceInConfigurableProcessTreeFitnessFunction(PluginContext context, XLog log, ProcessTree tree, 
			GeneticParameters parameters, GeneticAlgorithm problem){
		this.log = log;
		this.tree = tree;
		this.conformancePlugin = new PTQualityCalculator();
		this.context = context;
		this.problem = problem;
		this.parameters = parameters;
	}
	
	public double evaluate(IChromosome chromosome) {
		//System.out.println("chromosome : "+chromosome);
		//Apply configuration according to the chromosome
		double fitness=0.0 , overallFitness=0.0, numerator, denominator;
        double precision=0.0, simplicity=0.0, generalization=0.0, conformance=0.0, bestConformance=0.0;
        
        //final ConfigurableProcessesResult confProcResults = null;
		ProcessTree configuredProcessTree = problem.convertFromChromosomeToConfiguredProcessTree(chromosome);
		TreeFitness conf = conformancePlugin.calculatePTQuality(context, log, configuredProcessTree, XLogInfoImpl.NAME_CLASSIFIER);
    	TObjectDoubleIterator<TreeFitnessInfo> it = conf.fitnessValues.iterator();
		while (it.hasNext()) {
			it.advance();
			if (it.key().toString().equals("Fr"))
				fitness = it.value();
			if (it.key().toString().equals("Of"))
				overallFitness = it.value();
			if (it.key().toString().equals("Su"))
				simplicity = it.value();
			if (it.key().toString().equals("Pe"))
				precision = it.value();
			if (it.key().toString().equals("Gv"))
				generalization = it.value();
		}

		numerator = fitness*parameters.getFitnessWeigth()/((double) 100) + precision*parameters.getPrecisionWeigth()/((double) 100) + 
				  generalization*parameters.getGeneralizationWeigth()/((double) 100) + simplicity*parameters.getSimplicityWeigth()/((double) 100);
		denominator = (parameters.getFitnessWeigth()/((double) 100) + parameters.getPrecisionWeigth()/((double) 100) + parameters.getGeneralizationWeigth()/((double) 100) + parameters.getSimplicityWeigth()/((double) 100));

		conformance = numerator/denominator;
		//System.out.println("CONF ---- : "+conf);
		return conformance;
 	}
 

}
