package org.processmininguc.configurableprocesses.plugins;

import java.text.DecimalFormat;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmininguc.configurableprocesses.algorithms.ConfigurableProcessesResult;

@Plugin(name = "Configured Process Tree Visualizer", 
		returnLabels = { "COnfigured Process Tree Visualized" }, 
		returnTypes = { JComponent.class }, 
		parameterLabels = { "ConfigurableProcessesResult" }, 
		userAccessible = true)
@Visualizer
public class ConfiguredProcessTreeVisualization {
	@PluginVariant(requiredParameterLabels = { 0 })
	public JComponent visualize(UIPluginContext context, ConfigurableProcessesResult result) {
		
		JTextPane body = new JTextPane();
		body.setBorder(BorderFactory.createEmptyBorder());
		body.setContentType("text/html");
		body.setEditable(false);
		body.setCaretPosition(0);
		body.setText(toHTMLString(true,result));
		JScrollPane jsp = new JScrollPane(body);
		return jsp;
	}
	
	public String toHTMLString(boolean includeHTMLTags, ConfigurableProcessesResult results) {
		StringBuffer buffer = new StringBuffer();
		if (includeHTMLTags) {
			buffer.append("<html>");
		}
		
		buffer.append("<table style=\"border:1px solid black;font-size:18px\">");

		// Time elapsed
		buffer.append("<tr>");
			buffer.append("<td  style=\"border:1px solid black;\" colspan=\"2\">");
			buffer.append("Time Elapsed");
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\" colspan=\"2\">");
			buffer.append(results.getTimeElapsed());
			buffer.append("</td>");
		buffer.append("</tr>");

		// Conformance metric
		buffer.append("<tr>");
			buffer.append("<td  style=\"border:1px solid black;\" colspan=\"2\">");
			buffer.append("Conformance");
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\" colspan=\"2\">");
			buffer.append(new DecimalFormat("#.###").format(results.getConformance()));
			buffer.append("</td>");
		buffer.append("</tr>");

		// Conformance metric fitness
		buffer.append("<tr>");
			buffer.append("<td rowspan=\"4\">");
			buffer.append("quality metrics");
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append("fitness");
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append(new DecimalFormat("#.###").format(results.getFitness()));
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append(new DecimalFormat("#.###").format(results.getParamFitness()) + " %");
			buffer.append("</td>");
		buffer.append("</tr>");

		// Conformance metric Precision
		buffer.append("<tr>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append("precision");
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append(new DecimalFormat("#.###").format(results.getPrecision()) );
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append(new DecimalFormat("#.###").format(results.getParamPrecision()) + " %");
			buffer.append("</td>");
		buffer.append("</tr>");

		// Conformance metric Generalization
		buffer.append("<tr>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append("generalization");
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append(new DecimalFormat("#.###").format(results.getGeneralization()));
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append(new DecimalFormat("#.###").format(results.getParamGeneralization()) + " %");
			buffer.append("</td>");
		buffer.append("</tr>");

		// Conformance metric Simplicity
		buffer.append("<tr>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append("simplicity");
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append(new DecimalFormat("#.###").format(results.getSimplicity()));
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\">");
			buffer.append(new DecimalFormat("#.###").format(results.getParamSimplicity()) + " %");
			buffer.append("</td>");
		buffer.append("</tr>");

		//number of configuations
		buffer.append("<tr>");
			buffer.append("<td  style=\"border:1px solid black;\" colspan=\"2\">");
			buffer.append("Number of configurations");
			buffer.append("</td>");
			buffer.append("<td style=\"border:1px solid black;\" colspan=\"2\">");
			buffer.append(results.getConfigurationSize());
			buffer.append("</td>");
		buffer.append("</tr>");

		//configuration activity names
		buffer.append("<tr>");
			buffer.append("<td rowspan=\"" + (results.getFinalConf().size()+1) +"\">");
			buffer.append("Configured nodes");
			buffer.append("</td>");
		buffer.append("</tr>");
			
		for(Map.Entry<String, String> entry : results.getFinalConf().entrySet()){
			buffer.append("<tr>");
				buffer.append("<td width=\"350\"  style=\"border:1px solid black;\">");
				buffer.append(entry.getKey());
				buffer.append("</td>");
				buffer.append("<td style=\"border:1px solid black;\" colspan=\"2\">");
				buffer.append(entry.getValue());
				buffer.append("</td>");
			buffer.append("</tr>");
		}
			//buffer.append("<td style=\"border:1px solid black;\">");
			//buffer.append(new DecimalFormat("#.###").format(results.getParamFitness()) + " %");
			//buffer.append("</td>");
		buffer.append("</tr>");
		
		// Overall conformance metric given by TreeFitness class
/*		buffer.append("<tr>");
			buffer.append("<td>");
			buffer.append("Overall Conformance");
			buffer.append("</td>");
			buffer.append("<td>");
			buffer.append(results.getOveralFitnessMetric());
			buffer.append("</td>");
		buffer.append("</tr>");
*/		
		
		buffer.append("</table>");
		buffer.append("</br>");
		
		if (includeHTMLTags) {
			buffer.append("</html>");
		}
		return buffer.toString();
	}
	
}
