/**
 * 
 */
package org.processmininguc.configurableprocesses.tests;

import gnu.trove.map.TObjectShortMap;
import gnu.trove.map.hash.TObjectShortHashMap;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.UUID;

import nl.tue.astar.AStarThread.Canceller;

import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.etm.CentralRegistry;
import org.processmining.plugins.etm.fitness.TreeFitness;
import org.processmining.plugins.etm.fitness.TreeFitnessAbstract;
import org.processmining.plugins.etm.fitness.metrics.FitnessReplay;
import org.processmining.plugins.etm.model.narytree.Configuration;
import org.processmining.plugins.etm.model.narytree.NAryTree;
import org.processmining.plugins.etm.model.narytree.Simulator;
import org.processmining.plugins.etm.model.narytree.TreeUtils;
import org.processmining.plugins.etm.model.narytree.conversion.NAryTreeToProcessTree;
import org.processmining.plugins.etm.model.narytree.conversion.ProcessTreeToNAryTree;
import org.processmining.plugins.etm.model.narytree.test.LogCreator;
import org.processmining.plugins.etm.model.narytree.test.Test;
import org.processmining.plugins.etm.parameters.ETMParamFactory;
import org.processmining.plugins.etm.ui.plugins.PTQualityCalculator;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.configuration.controlflow.ControlFlowConfiguration;
import org.processmining.processtree.configuration.controlflow.impl.ControlFlowConfigurationImpl;

/**
 * @author mauro
 *
 */
public class TreeTests{
	private static XEventClasses eventClasses;
	
	public static void evaluateBasicTree(){
		Configuration c1 = new Configuration(13);
		c1.conf[2]=Configuration.HIDDEN;
		c1.conf[12]=Configuration.HIDDEN;

		Configuration c2 = new Configuration(13);
		c2.conf[7]=Configuration.BLOCKED;
		c2.conf[12]=Configuration.HIDDEN;
		
		Configuration c3 = new Configuration(13);
		c3.conf[7]=Configuration.BLOCKED;
		c3.conf[12]=Configuration.HIDDEN;
		

		String[][] logString2 = new String[][] { { "A", "B", "C", "G", "T", "G", "T", "G", "Q", "Z" },
				{ "A", "C", "B", "G", "T", "G", "T", "G", "Q", "Z" }, { "A", "M", "G", "T", "G", "Q", "Z" },
				{ "A", "N", "G", "Q", "Z" }};
		XLog log = LogCreator.createLog(logString2, new int[] { 50, 50, 25, 35 });


		CentralRegistry registry = new CentralRegistry(log, XLogInfoImpl.NAME_CLASSIFIER, new Random());

		String treeStringFr = "SEQ( LEAF: A , XOR( AND( LEAF: B , LEAF: C ) , LEAF: M , LEAF: N ) , LOOP( LEAF: G , LEAF: T , LEAF: Q ) , LEAF: Z ) ";

		NAryTree treeFr = TreeUtils.fromString(treeStringFr, registry.getEventClasses());

		treeFr.addConfiguration(c1);
		treeFr.addConfiguration(c2);
		
		NAryTree res_c1 = treeFr.applyConfiguration(0);
		NAryTree res_c2 = treeFr.applyConfiguration(1);

		Canceller c = new Canceller() {
			public boolean isCancelled() {
				return false;
			}
		};
		
		ProcessTree convertedTree = null;
		
		TreeFitnessAbstract of = ETMParamFactory.createStandardOverallFitness(registry);
		System.out.println("get fitness tree 1: "+of.getFitness(res_c1, null));
		System.out.println(TreeUtils.toString(res_c1, registry.getEventClasses()));
		System.out.println(registry.getFitness(res_c1));
		System.out.println("Tree full info: " + treeFr.toInternalString());
		System.out.println("Tree : " + treeFr.toString());
		System.out.println();

		of.getFitness(res_c2, null);
		
//		System.out.println("get fitness tree 2: "+of.getFitness(res_c2, null));
//		System.out.println(TreeUtils.toString(res_c2, registry.getEventClasses()));
//		System.out.println(registry.getFitness(res_c2));
//		System.out.println("Tree full info: " + treeFr.toInternalString());
//		System.out.println("Tree : " + treeFr.toString());
//		System.out.println();

		System.out.println("Configurations : " + treeFr.getNumberOfConfigurations());
		System.out.println("Configuration 1 : " + treeFr.getConfiguration(0));
		System.out.println("Configuration 2 : " + treeFr.getConfiguration(1));
		//System.out.println("Configuration 2 : " + treeFr.getConfiguration(2));

		System.out.println("Fr conf-1: " + registry.getFitness(res_c1).fitnessValues.get(FitnessReplay.info));
		convertedTree =  NAryTreeToProcessTree.convert(res_c1, registry.getEventClasses());
		System.out.println("converted tree conf1: "+convertedTree);

//		registry.clearFitnessCache();
//		registry = new CentralRegistry(log, XLogInfoImpl.NAME_CLASSIFIER, new Random());
//		of = ETMParamFactory.createStandardOverallFitness(registry);
//		treeFr = TreeUtils.fromString(treeStringFr, registry.getEventClasses());
//		treeFr.addConfiguration(c2);
		//NAryTree res_c2 = treeFr.applyConfiguration(1);

		System.out.println("Fr conf-2: " + registry.getFitness(res_c2).fitnessValues.get(FitnessReplay.info));
		convertedTree =  NAryTreeToProcessTree.convert(res_c2, registry.getEventClasses());
		System.out.println("converted tree conf2: "+convertedTree);

		//System.out.println("Fr conf-3: " + registry.getFitness(res_c3).fitnessValues.get(FitnessReplay.info));
		//convertedTree =  NAryTreeToProcessTree.convert(res_c3, registry.getEventClasses());
		System.out.println("converted tree conf3: "+convertedTree);

		
		//Another way to add configurations
/*		System.out.println("***** Another way to add configurations *****");

		NAryTree treeNew = TreeUtils.fromString(treeStringFr);

		//Add configurations
		for (int index = 0; index < treeNew.nChildren(0); index++) {
			//Build configuration
			treeNew.addConfiguration(new Configuration(treeNew.size()));
		}
		System.out.println("tree confs : "+treeNew.getNumberOfConfigurations());
		System.out.println("toInternalString : "+treeNew.toInternalString());
		//For each child of the root
		for (int index = 0; index < treeNew.nChildren(0); index++) {
			//Block for al configurations except child index :)
			for (int i = 0; i < treeNew.nChildren(0); i++) {
				if (i != index) {
					int childIndex = treeNew.getChildAtIndex(0, index);
					treeNew.setNodeConfiguration(i, childIndex, Configuration.BLOCKED);
				}
			}
		}
		
		System.out.println(TreeUtils.toString(treeNew));
*/		
	}
	
	public static void evaluateBasicTreeMultConf(ProcessTree tree, XLog log){
		CentralRegistry registry = new CentralRegistry(log, XLogInfoImpl.NAME_CLASSIFIER, new Random());
		
		System.out.println("\n*** evaluateBasicTreeMultConf ***");
		System.out.println("tree name : " + tree.getName());
		System.out.println("tree : " + tree.toString());
		System.out.println("tree has configurations? : " + tree.isConfigurable());
		
		ControlFlowConfiguration controlFlowConfs = new ControlFlowConfigurationImpl();
		List<ControlFlowConfiguration> listconfs = new ArrayList<ControlFlowConfiguration>();

		for (Edge edge :tree.getEdges()){
			if (edge.isBlockable()){
				controlFlowConfs.addBlockedEdge(edge.getID());
			}	
			if (edge.isHideable()){
				controlFlowConfs.addHiddenEdge(edge.getID());
			}
		}
		
		for (UUID cfcBlocked :controlFlowConfs.getBlockedEdges()){
			for (UUID cfcHidden :controlFlowConfs.getHiddenEdges()){
				Node myuuid = tree.getEdge(cfcBlocked).getTarget();
				System.out.println("["+tree.getEdge(cfcBlocked).getTarget()+","+cfcHidden+"]");
			}
		}
		
		listconfs.add(controlFlowConfs);
	
		System.out.println("blocked edges : "+controlFlowConfs.getBlockedEdges());
		System.out.println("hidden edges : "+controlFlowConfs.getHiddenEdges());
		System.out.println("list of confs : "+ listconfs);
			
		ProcessTreeToNAryTree simpleUS = new ProcessTreeToNAryTree(registry.getEventClasses());
		NAryTree simpleNAryTree = simpleUS.convert(tree);

/*		//Alternative way to add configurations			
		NAryTree NAryTreeTest = simpleUS.convert(tree);
		NAryTreeTest.addConfiguration(new Configuration(NAryTreeTest.size()));
		NAryTreeTest.addConfiguration(new Configuration(NAryTreeTest.size()));		
		System.out.println("NAryTreeTest tree : "+NAryTreeTest);
		NAryTreeTest.setNodeConfiguration(0, 1, Configuration.BLOCKED);
		System.out.println("NAryTreeTest tree : "+NAryTreeTest);
		System.out.println("NAryTreeTest full info : "+NAryTreeTest.toInternalString());
*/		
		System.out.println("tree #conf : " + simpleNAryTree.getNumberOfConfigurations());
		System.out.println("Tree full info: " + simpleNAryTree.toInternalString());
		
		System.out.println("get type 3 : "+simpleNAryTree.getType(1));
		System.out.println("getEventClasses size : "+simpleUS.getEventClasses().size());
		System.out.println("getEventClasses classes : "+simpleUS.getEventClasses().getClasses());
		System.out.println("getEventClasses index 1 : "+simpleUS.getEventClasses().getByIndex(1));
		
		List<UUID> myuuids = new ArrayList<UUID>();
		
		Map<UUID,List<Byte>> uuidsConfsMap = new HashMap<UUID,List<Byte>>();
		List<List<Pair<UUID,Byte>>> pairList = new ArrayList<List<Pair<UUID,Byte>>>();
		
		for (Edge edge :tree.getEdges()){
			List<Byte> bytelist = new ArrayList<Byte>();
			if (edge.isBlockable()){
				bytelist.add(Configuration.BLOCKED);
			}
			if(edge.isHideable()){				
				bytelist.add(Configuration.HIDDEN);
			}
			if(edge.isBlockable() || edge.isHideable()){				
				bytelist.add(Configuration.NOTCONFIGURED);
				uuidsConfsMap.put(tree.getEdge(edge.getID()).getTarget().getID(),bytelist);
			}
		}
		
		//Create list of pairs<UUID,Byte>
    	Pair<UUID, Byte> p = new Pair(null,null);
        for (Map.Entry<UUID, List<Byte>> entry : uuidsConfsMap.entrySet()) {
        	List<Pair<UUID,Byte>> temp = new ArrayList<Pair<UUID,Byte>>();
            for (Byte ent : entry.getValue()){
            	p = new Pair(entry.getKey(),ent);
            	temp.add(p);
            	//System.out.println(entry.getKey()+" - "+ent);
            }
        	pairList.add(temp);
        }			
		System.out.println("Pair<UUID,Byte> : "+pairList);
        
		/*// iterate and display values
        System.out.println("Fetching Keys and corresponding [Multiple] Values \n");
        for (Map.Entry<UUID, List<Byte>> entry : uuidsConfsMap.entrySet()) {
            UUID key = entry.getKey();
            List<Byte> values = entry.getValue();
            System.out.println("Key = " + key);
            System.out.println("Values = " + values + "\n");
        }*/			
        
        //combineAllCases(uuidsConfsMap);
        List<List<Pair<UUID,Byte>>> cartProdList = new ArrayList<List<Pair<UUID,Byte>>>();
        //cartProdList = cartesianProduct(pairList);
        System.out.println("all comb : "+cartProdList);
        TreeFitnessAbstract of = null;
        ProcessTree convertedTree=null, bestTree=null;
        Configuration c=null, bestConfiguration=null;
        int i=0;
        double bestFitness=0.0, bestOverallFitness=0.0, overallFitness=0.0;
        
        for(List<Pair<UUID,Byte>> confList : cartProdList){
        	System.out.println("******* Proccessing configuration "+(i+1)+" *******");
        	c = new Configuration(simpleNAryTree.size());
        	for(Pair<UUID,Byte> comb : confList){ //cartProdList.get(0) o confList
        		if (comb.getR().equals(Configuration.BLOCKED)){
        			c.conf[myuuids.indexOf(comb.getL())]=Configuration.BLOCKED;
        		}else if(comb.getR().equals(Configuration.HIDDEN)){
        			c.conf[myuuids.indexOf(comb.getL())]=Configuration.HIDDEN;
        		}else if(comb.getR().equals(Configuration.NOTCONFIGURED)){
        			c.conf[myuuids.indexOf(comb.getL())]=Configuration.NOTCONFIGURED;
        		}          
        	}
        	simpleNAryTree.addConfiguration(c);
        	NAryTree res = simpleNAryTree.applyConfiguration(0);
    		of = ETMParamFactory.createStandardOverallFitness(registry);
    		overallFitness = of.getFitness(res, null);
        	System.out.println("Tree full info Confs: " + simpleNAryTree.toInternalString());
        	System.out.println("overall fitness : "+overallFitness);
    		System.out.println("fitness : "+registry.getFitness(res).fitnessValues.get(FitnessReplay.info));

    		convertedTree =  NAryTreeToProcessTree.convert(res, registry.getEventClasses());
    		System.out.println("converted tree : "+convertedTree);
        	i++;
        	simpleNAryTree.removeConfiguration(0);
        	if (registry.getFitness(res).fitnessValues.get(FitnessReplay.info) > bestFitness){
        		bestFitness = registry.getFitness(res).fitnessValues.get(FitnessReplay.info);
        		bestOverallFitness = overallFitness;
        		bestTree = NAryTreeToProcessTree.convert(res, registry.getEventClasses());
        		bestConfiguration = c;
        	}else if (registry.getFitness(res).fitnessValues.get(FitnessReplay.info) == bestFitness){
        		if (overallFitness > bestOverallFitness){
            		bestFitness = registry.getFitness(res).fitnessValues.get(FitnessReplay.info);
            		bestOverallFitness = overallFitness;        			
            		bestTree = NAryTreeToProcessTree.convert(res, registry.getEventClasses());
            		bestConfiguration = c;
        		}
        	}
        	
        }
        
        System.out.println("==== Show Best Configuration ====");
        System.out.println(" best Tree : "+bestTree+
        					"\n best Configuration : "+bestConfiguration+
        					"\n best fitness : "+bestFitness+
        					"\n best overall fitness : "+bestOverallFitness);
		
        //System.out.println("fist all comb : "+cartProdList.get(0).size());
        
	}
	
	public static void simpleConfigurationTest(ProcessTree tree, XLog log){
		CentralRegistry registry = new CentralRegistry(log, XLogInfoImpl.NAME_CLASSIFIER, new Random());
		ControlFlowConfiguration controlFlowConf = new ControlFlowConfigurationImpl();
		List<ControlFlowConfiguration> listconfs = new ArrayList<ControlFlowConfiguration>();
		List<Edge> edges = new ArrayList<Edge>(tree.getEdges());

		controlFlowConf.addHiddenEdge(edges.get(0).getID());
		listconfs.add(controlFlowConf);
		
		ProcessTreeToNAryTree converter = new ProcessTreeToNAryTree(registry.getEventClasses());
		NAryTree simpleNAryTree = converter.convert(tree,listconfs);
		
		System.out.println("tree : "+simpleNAryTree.toString());
		System.out.println("tree full info : "+simpleNAryTree.toInternalString());
	
	}
		
	public static void testConformanceExternalPlugin(PluginContext context, XLog log, ProcessTree tree){
		PTQualityCalculator PTQuality = new PTQualityCalculator();
		TreeFitness fitnessMetric = PTQuality.calculatePTQuality(context, log, tree, XLogInfoImpl.NAME_CLASSIFIER);
		double[] fitnessArray = fitnessMetric.fitnessValues.values();
		System.out.println("fitnessMetric values : "+fitnessMetric.fitnessValues);
		//{Gv=0.7999999999999999,Fr=1.0,Su=1.0,Pe=1.0,Of=0.9882352941176471}
		System.out.println("fitnessMetric values : "+fitnessArray[1]);
	}
	
	public static ProcessTree createRandomProcessTree(){
		final double noise = 0.0;
		final long logSeed = 5897630,treeSeed=34121;
		final int numTraces = 25;
		final Random logRandom = new Random(logSeed);
		final Random treeRandom = new Random(treeSeed);
		final int size = 20;
		//NAryTree naryTree = TreeUtils.randomTree(10, .3, 4, 8);
		//TreeUtils.printTree(naryTree);
		String[] activities = new String[] { "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N" };

		TObjectShortMap<String> map = new TObjectShortHashMap<String>();
		// initialize a tree with all node costs 5 for leafs and 0 otherwise.
		for (short i = 0; i < activities.length; i++) {
			map.put(activities[i], i);
		}
		
		//ProcessTree processTree = NAryTreeToProcessTree.convert(naryTree);
		//List<Node> listNodes = new ArrayList<Node>(processTree.getNodes());
		//List<String> activityList = new ArrayList<String>();
		//for (Node node : listNodes){
		//	if (node.isLeaf())
		//		activityList.add(node.getName());
		//}
		//String[] activities = activityList.toArray(new String[activityList.size()]);
		//TObjectShortMap<String> map = new TObjectShortHashMap<String>();
		//for (short i = 0; i < activities.length; i++) {
		//	map.put(activities[i], i);
		//}
				
		//System.out.println("Activity List : " + activities.length);
		NAryTree tree = Test.getTree(map, treeRandom, size);
		System.out.println("tree : " + tree.toInternalString());
		Simulator sim = new Simulator(tree, 0, new Random(5897630));
		//System.out.println("sim : " + sim.toString());
		//String[][] traces = PerformanceEval.getTraces(sim, numTraces, activities, noise);
		String[][] traces = new String[numTraces][];
		traces[0] = new String[] {};
		for (int i = 1; i < numTraces; i++) {
			traces[i] = sim.getRandomTrace(activities, noise);
			System.out.println("traces["+i+"] : " + traces[i]);
		}
		
/*		XLog log;
		XFactoryRegistry.instance().setCurrentDefault(new XFactoryNaiveImpl());
		log = LogCreator.createLog(traces);
		System.out.println("traces : " + traces); 
*/		ProcessTree processTree = NAryTreeToProcessTree.convert(tree);
		return processTree;
	}
	
	public static void main(String[] args) {
		//evaluateBasicTree();
		//evaluateBasicTreeMultConf(tree, log);
		//impleConfigurationTest(tree,log);
		//testConformanceExternalPlugin(context, log,  tree);
		createRandomProcessTree();
	}
	
}
